/**
 * WildCAT: A Generic Framework for Context-Aware Applications.
 * Copyright (C) 2008 Bull S.A.S.
 * Copyright (C) 2008 EMN
 * Contact: wildcat@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors:
 *		- Pierre-Charles David (initial)
 *		- Nicolas Loriant
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.ow2.wildcat.hierarchy;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import javax.jms.Topic;
import javax.jms.TopicConnectionFactory;

import net.esper.client.Configuration;
import net.esper.client.EPException;
import net.esper.client.EPServiceProvider;
import net.esper.client.EPServiceProviderManager;
import net.esper.client.EPStatement;
import net.esper.client.UpdateListener;
import net.esper.event.EventBean;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.cmi.info.ClusteredObjectInfo;
import org.ow2.cmi.jndi.ClusteredObject;
import org.ow2.wildcat.Context;
import org.ow2.wildcat.ContextException;
import org.ow2.wildcat.ContextFactory;
import org.ow2.wildcat.ContextFactoryException;
import org.ow2.wildcat.Query;
import org.ow2.wildcat.event.WAttributeEvent;
import org.ow2.wildcat.event.WEvent;
import org.ow2.wildcat.hierarchy.attribute.Attribute;
import org.ow2.wildcat.hierarchy.attribute.POJOAttribute;
import org.ow2.wildcat.hierarchy.attribute.RunnableAttribute;
import org.ow2.wildcat.hierarchy.resource.LocalResource;
import org.ow2.wildcat.hierarchy.resource.Resource;
import org.ow2.wildcat.hierarchy.resource.ResourceException;
import org.ow2.wildcat.hierarchy.resource.Symlink;
import org.ow2.wildcat.remote.RemoteContext;
import org.ow2.wildcat.remote.dispatcher.rmi.RMIDispatcher;
import org.ow2.wildcat.remote.dispatcher.rmi.RMIRemoteContext;
import org.ow2.wildcat.remote.provider.JMSProviderException;
import org.ow2.wildcat.util.Messages;

/**
 * Basic implementation of the {@link Context} interface.
 */
public class BasicContext implements Context, ClusteredObject, RMIRemoteContext {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(BasicContext.class);

    /**
     * Context's name as seen by remote contexts.
     */
    private String name;

    private ContextFactory factory;

    /**
     * The root resource.
     */
    private final Resource root;

    /**
     * The Esper engine.
     */
    private final EPServiceProvider esperServiceProvider;

    /**
     * Index of all EPStatement registered.
     */
    private final Map<Query, EPStatement> queries;

    private final ScheduledExecutorService scheduler =
        Executors.newScheduledThreadPool(1);

    private final ScheduledThreadPoolExecutor runnableExecutor =
        new ScheduledThreadPoolExecutor(10);

    private final Map<String, Future<?>> tasks =
        new HashMap<String, Future<?>>();

    private boolean exported;

    private Query forwarderQuery;

    private EventForwarder eventForwarder;

    private final Map<Path, AttributePoller> attributePollers;


    /**
     * Constructor.
     *
     * @param factory the associate factory
     */
    protected BasicContext(final ContextFactory factory) {
        this.factory = factory;
        this.exported = false;
        attributePollers = new HashMap<Path, AttributePoller>();
        this.root = new LocalResource();
        this.root.mount(this, Path.ROOT);

        /* Esper Configuration */
        Configuration esperConfiguration = new Configuration();
        esperConfiguration.addEventTypeAlias("WEvent",
        "org.ow2.wildcat.event.WEvent");
        esperConfiguration.addEventTypeAlias("WHierarchyEvent",
        "org.ow2.wildcat.event.WHierarchyEvent");
        esperConfiguration.addEventTypeAlias("WAttributeEvent",
        "org.ow2.wildcat.event.WAttributeEvent");
        this.esperServiceProvider = EPServiceProviderManager.getProvider(
                "wildcat", esperConfiguration);
        this.queries = new HashMap<Query, EPStatement>();
    }

    /**
     * Constructor.
     *
     * @param factory
     *            the associate factory
     * @param name
     *            context's name
     */
    public BasicContext(final ContextFactory factory, final String name) {
        this(factory);
        logger.debug(Messages.message(BasicContext.class.getName() + ".create",
                name));
        this.name = name;
    }

    /**
     * Constructor, instanciate a local-only context.
     *
     * @param name
     *            context's name
     */
    public BasicContext(final String name) {
        this(null, name);
    }

    /**
     * Constructor, instanciate a local-only context named "anonymous".
     */
    public BasicContext() {
        this("anonymous");
    }

    /*
     * PULL RELATED
     */

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#getName()
     */
    public String getName() {
        return this.name;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#attachAttribute(java.lang.String,
     *      org.ow2.wildcat.hierarchy.attribute.Attribute)
     */
    public void attachAttribute(final String attribute, final Attribute a) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".attachAttribute", attribute));

        Path attr = null;
        RemoteContext ctx = null;
        try {
            attr = new Path(attribute).canonicalize();
            attr.expectAbsolute();
            attr.expectAttribute();
            attr.expectFinite();

            ctx = this.resolve(attr.getHostname());
            if (ctx == this) {
                this.attachAttribute(attr, a);
                if(a instanceof RunnableAttribute) {
                    RunnableAttribute runnableAttribute = (RunnableAttribute) a;
                    if(runnableAttribute.getDelay() >= 0) {
                        addRunnable(name, runnableAttribute);
                    }
                }
            } else {
                try {
                    ctx.attachAttribute(attribute, a);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
    }

    /**
     * @param attribute
     * @return
     * @throws ContextException
     */
    public Attribute getAttribute(final String attribute) throws ContextException {
        Path attr = null;
        RemoteContext ctx = null;
        try {
            attr = new Path(attribute).canonicalize();
            attr.expectAbsolute();
            attr.expectAttribute();
            attr.expectFinite();

            ctx = this.resolve(attr.getHostname());
            if (ctx == this) {
                return this.getAttribute(attr);
            } else {
                try {
                    return ctx.getAttribute(attribute);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
            return null;
        }
    }

    /**
     * @param attr
     * @return
     * @throws ContextException
     */
    private Attribute getAttribute(final Path attr) throws ContextException {
        try {
            return this.root.getAttribute(attr);
        } catch (ResourceException e) {
            logger.error("Unable to get an attribute at path " + attr, e);
            throw new ContextException("Unable to get an attribute at path " + attr, e);
        }
    }

    private void attachAttribute(final Path attribute, final Attribute a) throws ContextException {
        try {
            this.root.attachAttribute(attribute.getRelativePart(), a);
        } catch (ResourceException e) {
            logger.error("Unable to attach an attribute at path " + attribute, e);
            throw new ContextException("Unable to attach an attribute at path " + attribute, e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#createAttribute(java.lang.String,
     *      java.lang.Object)
     */
    public void createAttribute(final String attribute, final Object value) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".createAttribute", attribute));

        Path attr = null;
        RemoteContext ctx = null;
        try {
            attr = new Path(attribute).canonicalize();
            attr.expectAbsolute();
            attr.expectAttribute();
            attr.expectFinite();

            ctx = this.resolve(attr.getHostname());
            if (ctx == this) {
                this.createAttribute(attr, value);
            } else {
                try {
                    ctx.createAttribute(attribute, value);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
    }

    public void createAttribute(final Path attribute, final Object value) throws ContextException {
        try {
            this.root.createAttribute(attribute.getRelativePart(), value);
        } catch (ResourceException e) {
            logger.error("Unable to create an attribute at path " + attribute, e);
            throw new ContextException("Unable to create an attribute at path " + attribute, e);
        }
    }

    public Attribute createAttribute(final Object value) {
        return new POJOAttribute(value);
    }

    /**
     * @param attribute
     * @throws ContextException
     */
    public void detachAttribute(final String attribute) throws ContextException {
        Path res = null;
        RemoteContext ctx;
        try {
            res = new Path(attribute).canonicalize();
            res.expectAbsolute();
            res.expectAttribute();
            res.expectFinite();

            ctx = this.resolve(res.getHostname());
            if (ctx == this) {
                Attribute att = getAttribute(res);
                if(att instanceof RunnableAttribute) {
                    removeRunnable(attribute);
                }
                this.detachAttribute(res);
            } else {
                try {
                    ctx.detachAttribute(attribute);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
    }


    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#attachResource(java.lang.String,
     *      org.ow2.wildcat.hierarchy.resource.Resource)
     */
    public void attachResource(final String resource, final Resource res) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".attachResource", resource));

        Path r = null;
        RemoteContext ctx = null;
        try {
            r = new Path(resource).canonicalize();
            r.expectAbsolute();
            r.expectResource();
            r.expectFinite();

            ctx = this.resolve(r.getHostname());
            if (ctx == this) {
                this.attachResource(r, res);
            } else {
                try {
                    ctx.attachResource(resource, res);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#createResource(java.lang.String)
     */
    public void createResource(final String resource) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".createResource", resource));

        Path res = null;
        RemoteContext ctx;
        try {
            res = new Path(resource).canonicalize();
            res.expectAbsolute();
            res.expectResource();
            res.expectFinite();

            ctx = this.resolve(res.getHostname());
            if (ctx == this) {
                this.createResource(res);
            } else {
                try {
                    ctx.createResource(resource);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
    }

    /**
     * @param resource
     * @return
     * @throws ContextException
     */
    public Resource getResource(final String resource) throws ContextException {
        Path res = null;
        RemoteContext ctx;
        try {
            res = new Path(resource).canonicalize();
            res.expectAbsolute();
            res.expectResource();
            res.expectFinite();

            ctx = this.resolve(res.getHostname());
            if (ctx == this) {
                return getResource(res);
            } else {
                try {
                    return ctx.getResource(resource);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
        return null;
    }

    /**
     * @param resource
     * @throws ContextException
     */
    public void detachResource(final String resource) throws ContextException {
        Path r = null;
        RemoteContext ctx = null;
        try {
            r = new Path(resource).canonicalize();
            r.expectAbsolute();
            r.expectResource();
            r.expectFinite();

            ctx = this.resolve(r.getHostname());
            if (ctx == this) {
                Resource symlink = getResource(r);
                if (symlink != null) {
                    if (symlink instanceof Symlink) {
                        try {
                            factory.doNotListen(this, ((Symlink) symlink).getTarget().getHostname());
                        } catch (ContextFactoryException e) {
                            logger.warn("Unable to not any more listen the context with name "
                                    + ((Symlink) symlink).getTarget().getHostname(), e);
                        }
                    }
                }
                this.detachResource(r);
            } else {
                try {
                    ctx.detachResource(resource);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
    }

    private void attachResource(final Path resource, final Resource res) throws ContextException {
        try {
            this.root.attachResource(resource.getRelativePart(), res);
        } catch (ResourceException e) {
            logger.error("Unable to attach a resource at path " + resource, e);
            throw new ContextException("Unable to attach a resource at path " + resource, e);
        }
    }

    /**
     * @param r
     * @throws ContextException
     */
    private void detachResource(final Path r) throws ContextException {
        try {
            this.root.detachResource(r.getRelativePart());
        } catch (ResourceException e) {
            logger.error("Unable to detach a resource at path " + r, e);
            throw new ContextException("Unable to detach a resource at path " + r, e);
        }
    }

    private void createResource(final Path resource) throws ContextException {
        try {
            this.root.createResource(resource.getRelativePart());
        } catch (ResourceException e) {
            logger.error("Unable to create a resource at path " + resource, e);
            throw new ContextException("Unable to create a resource at path " + resource, e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#createSymlink(java.lang.String,
     *      java.lang.String)
     */
    public void createSymlink(final String link, final String target) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".createSymlink", link, target));

        Path l = null;
        Path t = null;
        RemoteContext ctx = null;

        try {
            l = new Path(link).canonicalize();
            t = new Path(target).canonicalize();
            l.expectAbsolute();
            l.expectResource();
            l.expectFinite();
            t.expectAbsolute();
            t.expectResource();
            t.expectFinite();

            ctx = this.resolve(l.getHostname());
            if (ctx == this) {
                if (!this.isLocalHost(t.getHostname()) && this.factory != null) {
                    if(!this.factory.listen(this, t.getHostname())) {
                        logger.error("Unable to listen the context with name " + t.getHostname());
                        throw new ContextException("Unable to listen the context with name " + t.getHostname());
                    }
                }
                Resource sl = new Symlink(t);
                this.attachResource(l, sl);
            } else {
                try {
                    ctx.createSymlink(link, target);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#getValue(java.lang.String)
     */
    public Object getValue(final String attribute) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName() + ".pull",
                attribute));

        Path attr = null;
        RemoteContext ctx = null;
        try {
            attr = new Path(attribute).canonicalize();
            attr.expectAbsolute();
            attr.expectAttribute();
            attr.expectFinite();

            ctx = this.resolve(attr.getHostname());
            if (ctx == this) {
                return this.getValue(attr);
            } else {
                try {
                    return ctx.getValue(attribute);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
        }
        return null;
    }

    private Object getValue(final Path attribute) throws ContextException {
        try {
            return this.root.getValue(attribute.getRelativePart());
        } catch (ResourceException e) {
            logger.error("Unable to get a value at path " + attribute, e);
            throw new ContextException("Unable to get a value at path " + attribute, e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#setValue(java.lang.String, java.lang.Object)
     */
    public Object setValue(final String attribute, final Object value) throws ContextException {

        logger.debug(Messages.message(BasicContext.class.getName() + ".set",
                attribute, value));

        Path attr = null;
        RemoteContext ctx = null;
        try {
            attr = new Path(attribute).canonicalize();
            attr.expectAbsolute();
            attr.expectAttribute();
            attr.expectFinite();

            ctx = this.resolve(attr.getHostname());
            if (ctx == this) {
                return this.setValue(attr, value);
            } else {
                try {
                    return ctx.setValue(attribute, value);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
            return null;
        }
    }

    private Object setValue(final Path attribute, final Object value) throws ContextException {
        try {
            return this.root.setValue(attribute.getRelativePart(), value);
        } catch (ResourceException e) {
            logger.error("Unable to set a value at path " + attribute, e);
            throw new ContextException("Unable to set a value at path " + attribute, e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#list(java.lang.String)
     */
    public Set<String> list(final String resource) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName() + ".list"));

        Path res = null;
        RemoteContext ctx = null;
        try {
            res = new Path(resource).canonicalize();
            res.expectAbsolute();
            res.expectResource();
            res.expectFinite();

            ctx = this.resolve(res.getHostname());
            if (ctx == this) {
                return this.list(res);
            } else {
                try {
                    return ctx.list(resource);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
            return null;
        }
    }

    private Set<String> list(final Path resource) throws ContextException {
        try {
            return this.root.list(resource.getRelativePart());
        } catch (ResourceException e) {
            logger.error("Unable to list a resource at path " + resource, e);
            throw new ContextException("Unable to list a resource at path " + resource, e);
        }
    }

    /**
     * @param res
     * @return
     * @throws ContextException
     */
    private Resource getResource(final Path res) throws ContextException {
        try {
            return root.getResource(res);
        } catch (ResourceException e) {
            logger.error("Unable to get a resource at path " + res, e);
            throw new ContextException("Unable to get a resource at path " + res, e);
        }
    }

    /**
     * @param attribute
     * @throws ContextException
     */
    private void detachAttribute(final Path path) throws ContextException {
        try {
            this.root.detachAttribute(path);
        } catch (ResourceException e) {
            logger.error("Unable to detach an attribute at path " + path, e);
            throw new ContextException("Unable to detach an attribute at path " + path, e);
        }
    }

    /*
     * PUSH RELATED
     */

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#emitEvent(org.ow2.wildcat.event.PathEvent)
     */
    public void emitEvent(final WEvent... events) {
        for (Object event : events) {
            this.esperServiceProvider.getEPRuntime().sendEvent(event);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#createPeriodicAttributePoller(java.lang.String,
     *      java.lang.String)
     */
    public void createPeriodicAttributePoller(final String attribute, final long duration,
            final TimeUnit unit) throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".periodicPoll", attribute, duration, unit));

        Path attr = null;
        RemoteContext ctx = null;
        try {
            attr = new Path(attribute).canonicalize();
            attr.expectAbsolute();
            attr.expectAttribute();
            attr.expectFinite();

            ctx = this.resolve(attr.getHostname());
            if (ctx == this) {
                this.createPeriodicAttributePoller(attr, duration, unit);
            } else {
                try {
                    ctx.createPeriodicAttributePoller(attribute, duration, unit);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
            return;
        }
    }

    public void removePeriodicAttributePoller(final String attribute) throws ContextException {
        Path attr = null;
        RemoteContext ctx = null;
        try {
            attr = new Path(attribute).canonicalize();
            attr.expectAbsolute();
            attr.expectAttribute();
            attr.expectFinite();

            ctx = this.resolve(attr.getHostname());
            if (ctx == this) {
                this.removePeriodicAttributePoller(attr);
            } else {
                try {
                    ctx.removePeriodicAttributePoller(attribute);
                } catch (RemoteException e) {
                    logger.error("Communication error.", e);
                    throw new ContextException("Communication error.", e);
                }
            }
        } catch (MalformedPathException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"), e);
            return;
        }
    }

    private void createPeriodicAttributePoller(final Path attribute, final long duration,
            final TimeUnit period) {
        AttributePoller attributePoller = new AttributePoller(this, attribute, duration, period);
        attributePoller.start();
        attributePollers.put(attribute, attributePoller);
    }

    /**
     * @param attr
     */
    private void removePeriodicAttributePoller(final Path attr) {
        AttributePoller attributePoller = attributePollers.remove(attr);
        attributePoller.stop();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#createQuery(java.lang.String)
     */
    public Query createQuery(final String query) {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".createQuery", query));

        EPStatement statement = null;
        Query q = null;
        try {
            statement = this.esperServiceProvider.getEPAdministrator()
            .createEQL(query);
            q = new Query(query);
            this.queries.put(q, statement);
        } catch (EPException e) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".errorQuery", query), e);
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"));
        }
        return q;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#pauseQuery(org.ow2.wildcat.Query)
     */
    public void pauseQuery(final Query query) {
        EPStatement statement = this.queries.get(query);
        if (statement != null) {
            statement.stop();
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#resumeQuery(org.ow2.wildcat.Query)
     */
    public void resumeQuery(final Query query) {
        EPStatement statement = this.queries.get(query);
        if (statement != null) {
            statement.start();
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#destroyQuery(org.ow2.wildcat.Query)
     */
    public void destroyQuery(final Query query) {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".destroyQuery", query));

        if (!this.queries.containsKey(query)) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".noSuchQuery", query));
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"));
        } else {
            EPStatement statement = this.queries.remove(query);
            statement.destroy();
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#registerListener(java.lang.String,
     *      net.esper.client.UpdateListener)
     */
    public Query registerListeners(final String query, final UpdateListener... listeners) {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".createQuery", query));

        Query q = createQuery(query);
        if (q != null && listeners != null) {
            this.registerListeners(q, listeners);
        }
        return q;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#registerListener(org.ow2.wildcat.Query,
     *      net.esper.client.UpdateListener)
     */
    public void registerListeners(final Query query, final UpdateListener... listeners) {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".registerListener", query));

        if (!this.queries.containsKey(query)) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".noSuchQuery", query));
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"));
        } else {
            EPStatement statement = this.queries.get(query);
            if (listeners != null) {
                for (UpdateListener l : listeners) {
                    statement.addListener(l);
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#removeListeners(org.ow2.wildcat.Query,
     *      net.esper.client.UpdateListener[])
     */
    public void removeListeners(final Query query, final UpdateListener... listeners) {
        logger.debug(Messages.message(BasicContext.class.getName()
                + ".removeListener", query, listeners));

        if (!this.queries.containsKey(query)) {
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".noSuchQuery", query));
            logger.warn(Messages.message(BasicContext.class.getName()
                    + ".ignore"));
        } else {
            EPStatement statement = this.queries.get(query);
            if (listeners != null) {
                for (UpdateListener l : listeners) {
                    statement.removeListener(l);
                }
            }
        }
    }

    /*
     * HELPER
     */

    /**
     * @param hostname
     * @return <b>true</b> if hostname is null, or equals "self" or is current
     *         context name
     */
    private boolean isLocalHost(final String hostname) {
        return hostname == null || "self".equals(hostname)
        || hostname.equals(this.name);
    }

    /**
     * Gets a Context instance by its name.
     *
     * @param hostname
     *            name of the context to resolve
     * @return a context instance FW: to be used to resolve remote context,
     *         technology to be determined
     * @throws ContextException
     */
    private RemoteContext resolve(final String hostname) throws ContextException {
        try {
            return this.isLocalHost(hostname) ? this : this.factory
                    .lookup(hostname);
        } catch (ContextFactoryException e) {
            logger.error("Unable to resolve the context with name " + name, e);
            throw new ContextException("Unable to resolve the context with name " + name, e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.wildcat.Context#export()
     */
    public void export() throws ContextException {
        logger.debug(Messages.message(BasicContext.class.getName() + ".export",
                this.getName()));

        if (!exported) {
            if (this.factory == null) {
                this.factory = ContextFactory.getDefaultFactory();
            }
            try {
                this.factory.export(this);
            } catch (ContextFactoryException e) {
                logger.error("Unable to export the context with name " + name, e);
                throw new ContextException("Unable to export the context with name " + name, e);
            }
            this.eventForwarder = new EventForwarder(this);
            this.forwarderQuery = this.registerListeners(
                    "select * from WEvent", this.eventForwarder);
            this.exported = true;
        }
    }

    /**
     * @throws ContextException
     */
    public void unexport() throws ContextException {
        if(exported) {
            exported = false;
            if (this.factory == null) {
                this.factory = ContextFactory.getDefaultFactory();
            }
            try {
                this.factory.unexport(this);
            } catch (ContextFactoryException e) {
                logger.error("Unable to unexport the context with name " + name, e);
                throw new ContextException("Unable to unexport the context with name " + name, e);
            }
            removeListeners(forwarderQuery, eventForwarder);
        }
    }

    private class AttributePoller implements Runnable {
        BasicContext ctx;
        Path attribute;
        long period;
        TimeUnit unit;
        ScheduledFuture<?> future;

        /**
         * @param ctx
         * @param attribute
         * @param duration
         * @param period
         */
        private AttributePoller(final BasicContext ctx, final Path attribute, final long period,
                final TimeUnit unit) {
            super();
            this.ctx = ctx;
            this.attribute = attribute;
            this.period = period;
            this.unit = unit;
        }

        /*
         * (non-Javadoc)
         *
         * @see java.lang.Runnable#run()
         */
        public void run() {
            Object value;
            try {
                value = ctx.getValue(this.attribute.toString());
                ctx.emitEvent(new WAttributeEvent(this.attribute, value));
            } catch (ContextException e) {
                logger.warn("Enable to emit a new event for the attribute at path " + attribute, e);
            }
        }

        public synchronized void start() {
            if (future == null) {
                this.future = ctx.scheduler.scheduleAtFixedRate(this, 0,
                        period, unit);
            }
        }

        public synchronized void stop() {
            if (future != null) {
                this.future.cancel(false);
                future = null;
            }
        }
    }

    private class EventForwarder implements UpdateListener {
        private BasicContext context;

        public EventForwarder(final BasicContext context) {
            this.context = context;
        }

        /*
         * (non-Javadoc)
         *
         * @see net.esper.client.UpdateListener#update(net.esper.event.EventBean[],
         *      net.esper.event.EventBean[])
         */
        public void update(final EventBean[] arg0, final EventBean[] arg1) {
            for (EventBean bean : arg0) {
                WEvent wevent = (WEvent) bean.getUnderlying();
                Path p = null;
                try {
                    p = new Path(wevent.getSource());
                    if (this.context.isLocalHost(p.getHostname())) {
                        p.setHostName(this.context.getName());
                        WEvent evt = (WEvent) wevent.clone();
                        evt.setSource(p);
                        this.context.factory.notifyEvent(this.context, evt);
                    }
                } catch (Exception e) {
                    logger.warn("Unable to forward the event " + wevent, e);
                }
            }
        }
    }

    public ClusteredObjectInfo getClusteredObjectInfo() {
        return new ClusteredObjectInfo(RMIRemoteContext.class, "wild_cluster");
    }

    /**
     * @return
     * @throws RemoteException
     * @throws JMSProviderException
     */
    public TopicConnectionFactory getTopicConnectionFactoryToSpy() throws RemoteException, ContextException{
        try {
            return ((RMIDispatcher) factory.getDispatcher()).getRemoteTopicConnectionFactory();
        } catch (Exception e) {
            logger.error("Unable to get the topic connection factory of the context with name " + name, e);
            throw new ContextException("Unable to get the topic connection factory of the context with name " + name, e);
        }
    }

    /**
     * @return
     * @throws RemoteException
     * @throws ContextException
     */
    public Topic getTopicToSpy() throws RemoteException, ContextException {
        try {
            return ((RMIDispatcher) factory.getDispatcher()).getExportedTopic(name);
        } catch (Exception e) {
            logger.error("Unable to get the topic of the context with name " + name, e);
            throw new ContextException("Unable to get the topic of the context with name " + name, e);
        }
    }

    private void addRunnable(final String name, final RunnableAttribute runnableAttribute) {
        tasks.put(name, runnableExecutor.schedule(
                runnableAttribute, runnableAttribute.getDelay(), TimeUnit.MILLISECONDS));
    }

    private void removeRunnable(final String name) {
        Future<?> task = tasks.get(name);
        if (!task.isDone() && !task.cancel(true)) {
            logger.warn("Cannot cancel the task for the RunnableAttribute with name " + name);
        } else {
            runnableExecutor.purge();
        }
    }

    /**
     * @throws ContextException
     */
    public void destroy() throws ContextException {
        unexport();
        esperServiceProvider.destroy();
    }

    /**
     * @param source
     * @return
     */
    public boolean listen(final String source) {
        return factory.listen(this, source);
    }

    /**
     * @param context a context name
     * @return true if a remote context with the given name is available
     */
    public boolean isRemoteContextAvailable(final String context) {
        return factory.isRemoteContextAvailable(context);
    }

}
