/**
 * WildCAT: A Generic Framework for Context-Aware Applications.
 * Copyright (C) 2008 Bull S.A.S.
 * Copyright (C) 2008 EMN
 * Contact: wildcat@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * --------------------------------------------------------------------------
 * $Id: RemoteContext.java 350 2008-11-10 09:41:14Z loris $
 * --------------------------------------------------------------------------
 */
package org.ow2.wildcat.remote;

import java.rmi.RemoteException;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import org.ow2.wildcat.Context;
import org.ow2.wildcat.ContextException;
import org.ow2.wildcat.hierarchy.attribute.Attribute;
import org.ow2.wildcat.hierarchy.resource.Resource;

/**
 * Define the operations of a {@link Context that are remotely callable.
 * @author Loris Bouzonnet
 */
public interface RemoteContext {

    /**
     * @return the name of the context as registered for remote connection
     */
    String getName() throws RemoteException;

    /*
     * PULL RELATED OPERATIONS
     */

    /**
     * Attach Attribute at specified location.
     *
     * @param attribute
     *            path of the Attribute
     * @param attr
     *            an Attribute
     */
    void attachAttribute(String attribute, Attribute attr) throws RemoteException, ContextException;

    /**
     * Creates Resource along a path
     *
     * @param resource
     *            path to create
     */
    void createResource(String resource) throws RemoteException, ContextException;

    void detachResource(String resource) throws RemoteException, ContextException;

    /**
     * Attach Resource at specified location.
     *
     * @param resource
     *            path of the Resource
     * @param res
     *            a Resource
     */
    void attachResource(String resource, Resource res) throws RemoteException, ContextException;

    /**
     * @param resource
     * @return
     */
    Resource getResource(String resource) throws RemoteException, ContextException;

    /**
     * Creates a new attribute.
     *
     * @param attribute
     *            path of the attribute
     * @param value
     *            initial value of the attribute
     */
    void createAttribute(String attribute, Object value) throws RemoteException, ContextException;

    /**
     * @param attribute
     */
    public void detachAttribute(String attribute) throws RemoteException, ContextException;

    /**
     * @param attribute
     * @return
     */
    Attribute getAttribute(String attribute) throws RemoteException, ContextException;

    /**
     * Creates and attach a Symlink.
     *
     * @param link
     *            attach point
     * @param target
     *            target path of the symbolic link
     */
    void createSymlink(String link, String target) throws RemoteException, ContextException;

    /**
     * Fetchs attribute value.
     *
     * @param attribute
     *            the path of the attribute
     * @return the value of the attribute, or <b>null</b> if attribute does not
     *         exist
     */
    Object getValue(String attribute) throws RemoteException, ContextException;

    /**
     * Set Attribute's value.
     *
     * @param attribute
     *            path of the Attribute
     * @param value
     *            new value of the Attribute
     * @return previous value of the Attribute
     */
    Object setValue(String attribute, Object value) throws RemoteException, ContextException;

    /**
     * List resources and attributes.
     *
     * @param resource
     *            the resource to inspect
     * @return A set of String, attribute's names preceded with '#'
     */
    Set<String> list(String resource) throws RemoteException, ContextException;

    /**
     * Periodically creates an AttributeChangedEvent where old and new values
     * are equal.
     *
     * @param attribute
     *            the location of the attribute to poll
     * @param duration
     *            the period
     * @param unit
     *            the period unit
     */
    void createPeriodicAttributePoller(String attribute, long duration,
            TimeUnit unit) throws RemoteException, ContextException;

    void removePeriodicAttributePoller(String attribute) throws RemoteException, ContextException;

}
