/**
 * WildCAT: A Generic Framework for Context-Aware Applications.
 * Copyright (C) 2008 Bull S.A.S.
 * Copyright (C) 2008 EMN
 * Contact: wildcat@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * --------------------------------------------------------------------------
 * $Id: AbstractDispatcher.java 359 2008-11-17 10:21:14Z loris $
 * --------------------------------------------------------------------------
 */
package org.ow2.wildcat.remote.dispatcher;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.jms.TopicSubscriber;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.wildcat.Context;
import org.ow2.wildcat.ContextFactory;
import org.ow2.wildcat.event.WEvent;
import org.ow2.wildcat.remote.ContextEvent;
import org.ow2.wildcat.remote.RemoteContext;
import org.ow2.wildcat.remote.provider.JMSProvider;
import org.ow2.wildcat.remote.provider.JNDIProvider;

/**
 * An abstract implementation of a {@link IDispatcher}.
 * @author Loris Bouzonnet
 */
public abstract class AbstractDispatcher implements IDispatcher, MessageListener {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(AbstractDispatcher.class);

    /**
     * The {@link ContextFactory} that has created this {@link IDispatcher}.
     */
    private final ContextFactory factory;

    /**
     * Delegate to use to perform some JMS applications.
     */
    private JMSProvider jmsProvider;

    /**
     * Session to use to publish.
     */
    private final TopicSession sessionToPublish;

    /**
     * Topic connection (can be local).
     */
    private final TopicConnection topicConnection;

    /**
     * Contexts that are exported with this dispatcher.
     */
    private final Map<String, Export> exports;

    /**
     * Local contexts listening for events coming from a given context name.
     */
    private final Map<String, TopicToSpy> listeners;

    /**
     * Cache of the already accessed {@link RemoteContext}.
     */
    private Map<String, RemoteContext> remoteContextsCache;


    /**
     * Create a new dispatcher.
     * @param factory the context factory used to create this dispatcher
     * @param local true if the registry is local
     * @throws Exception
     */
    protected AbstractDispatcher(final ContextFactory factory, final boolean local) throws DispatcherException {
        this.factory = factory;
        this.exports = new HashMap<String, Export>();
        this.listeners = new HashMap<String, TopicToSpy>();
        this.remoteContextsCache = new HashMap<String, RemoteContext>();

        jmsProvider = new JNDIProvider(factory.getJMSProviderDelegate(), local);

        try {
            jmsProvider.init(factory);
            TopicConnectionFactory connectionFactory;
            connectionFactory = jmsProvider.getTopicConnectionFactory();
            topicConnection = connectionFactory.createTopicConnection();
            topicConnection.start();
            sessionToPublish = topicConnection.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
        } catch (Exception e) {
            logger.error("Unable to initialize the JMS connection", e);
            throw new DispatcherException("Unable to initialize the JMS connection", e);
        }
    }


    /**
     * Makes a context callable remotely.
     * @param context a context
     * @throws DispatcherException
     */
    public final void export(final Context context) throws DispatcherException {
        if (!exports.containsKey(context.getName())) {
            // Compute the topic name
            String topicName = factory.computeTopicJNDIName(context.getName());
            try {
                Topic topicToPublish = jmsProvider.getTopic(topicName);
                TopicPublisher publisher = sessionToPublish.createPublisher(topicToPublish);
                exports.put(context.getName(),
                        new Export(context, topicToPublish, publisher));
            } catch (Exception e) {
                logger.error("Unable to retrieve the topic with name " + topicName, e);
                throw new DispatcherException("Unable to retrieve the topic with name " + topicName, e);
            }
            doExport(context);
            logger.info("The context " + context.getName() + " is now exported.");
        }
    }

    protected abstract void doExport(Context context) throws DispatcherException;

    /**
     * Make listening a given event source by a given local context.
     * @param listener a local context
     * @param source an event source
     * @throws DispatcherException
     */
    public synchronized final boolean listen(final Context listener, final String source) {

        TopicToSpy ls = listeners.get(source);
        if (ls == null) {
            Topic topicToListen;
            TopicConnection connectionToListen;
            try {
                topicToListen = getTopicToListen(source);
                connectionToListen = getTopicConnectionToListen(source);
                TopicSession sessionToListen =
                    connectionToListen.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
                TopicSubscriber consumer =
                    sessionToListen.createSubscriber(topicToListen);
                consumer.setMessageListener(this);
                ls = new TopicToSpy(topicToListen, consumer, sessionToListen);
            } catch (Exception e) {
                logger.error("Unable to retrieve the topic with name " + source, e);
                return false;
            }
            listeners.put(source, ls);
        }
        Set<Context> tts = ls.getSpyingContexts();
        if (!tts.contains(listener)) {
            tts.add(listener);
        } else {
            logger.debug("Context \'" + source + "\' already spied, ignoring");
        }
        return true;
    }

    public synchronized final void notifyEvent(final Context sender, final WEvent... events) {
        Export export = this.exports.get(sender.getName());
        if (export == null) {
            // this context is not exported ignore remote notification
        } else {
            try {
                Message message = sessionToPublish.createObjectMessage(
                        new ContextEvent(sender.getName(),
                                events));
                export.getPublisher().publish(message);
            } catch (JMSException e) {
                logger.warn("Unable to publish the event {0} coming from {1}",
                        Arrays.asList(events), sender.getName(), e);
            }
        }
    }

    /**
     * Gets a reference to a remote context.
     *
     * @param alias remote context's name
     * @return a RemoteContext instance
     * @throws DispatcherException
     */
    public synchronized final RemoteContext lookup(final String alias) throws DispatcherException {
        RemoteContext ctx = this.remoteContextsCache.get(alias);
        if (ctx == null) {
            ctx = getRemoteContext(alias);
            this.remoteContextsCache.put(alias, ctx);
        }
        return ctx;
    }

    public void onMessage(final Message message) {
        if (message instanceof ObjectMessage) {
            try {
                Object obj = ((ObjectMessage) message).getObject();
                if (obj instanceof ContextEvent) {
                    ContextEvent contextEvent = (ContextEvent) obj;
                    Set<Context> ls = getListeners(contextEvent.getSender());
                    if (ls != null) {
                        for (Context l : ls) {
                            l.emitEvent(contextEvent.getEvents());
                        }
                    }
                }
            } catch (Exception e) {
                logger.warn("pb: ", e);
            }
        } else {
            logger.warn("pb: not instance of ObjectMessage: {0}", message);
        }
    }


    /**
     * @return
     */
     public Topic getExportedTopic(final String name) {
        return exports.get(name).getTopic();
     }

     /**
      * @param me
      * @param source
      */
     public void doNotListen(final Context me, final String source) {
         TopicToSpy tts = listeners.get(source);
         Set<Context> spyingContexts = tts.getSpyingContexts();
         if (tts != null && spyingContexts.contains(me)) {
             spyingContexts.remove(me);
             if (spyingContexts.isEmpty()) {
                 try {
                     tts.getConsumer().close();
                 } catch (JMSException e) {
                     logger.warn("Unable to close the consumer for " + source, e);
                 }
                 try {
                     stopTopicConnection(source);
                 } catch (DispatcherException e) {
                     logger.warn("Unable to stop the connection for " + source, e);
                 }
                 listeners.remove(tts);
             }
         }
     }

     /**
      * @param context
      * @throws DispatcherException
      */
     public void unexport(final Context context) throws DispatcherException {
         doUnexport(context);
         Export export = exports.remove(context.getName());
         if (export != null) {
             try {
                 export.getPublisher().close();
                 jmsProvider.deleteTopic(factory.computeTopicJNDIName(export.getTopic().getTopicName()));
             } catch (Exception e) {
                 logger.error("Error while unexporting the context with name " + context.getName(), e);
                 throw new DispatcherException(
                         "Error while unexporting the context with name" + context.getName(), e);
             }
         }
     }

     public boolean isRemoteContextAvailable(final String context) {
         Topic topicToListen;
         TopicConnection connectionToListen;
         TopicSession sessionToListen = null;
         try {
             topicToListen = getTopicToListen(context);
             connectionToListen = getTopicConnectionToListen(context);
             sessionToListen =
                 connectionToListen.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
             TopicSubscriber consumer =
                 sessionToListen.createSubscriber(topicToListen);
             consumer.close();
             return true;
         } catch (Exception e) {
             logger.debug("Unable to create a subscriber for the context with name " + context, e);
             return false;
         } finally {
             if (sessionToListen != null) {
                 try {
                     sessionToListen.close();
                 } catch (JMSException e) {
                     logger.debug("Cannot close a session", e);
                 }
             }
         }
     }

     /**
      * @param context
      * @throws DispatcherException
      */
     protected abstract void doUnexport(Context context) throws DispatcherException;


     /**
      * @return the factory used to create this dispatcher.
      */
     protected final ContextFactory getFactory() {
         return factory;
     }

    /**
      * @param sender
      * @return
      */
     protected final Set<Context> getListeners(final String sender) {
         return listeners.get(sender).getSpyingContexts();
     }

     /**
      * @param name
      * @return
      */
     protected final Context getExportedContext(final String name) {
         return exports.get(name).getContext();
     }

     /**
      * @param source
      * @return
      * @throws DispatcherException
      */
     protected TopicConnection getTopicConnectionToListen(final String source) throws DispatcherException {
         return topicConnection;
     }

     protected void stopTopicConnection(final String source) throws DispatcherException {
         // Do nothing
     }

     /**
      * @param source
      * @return
      * @throws DispatcherException
      */
     protected Topic getTopicToListen(final String source) throws DispatcherException {
         String topicName = factory.computeTopicName(source);
         try {
             return sessionToPublish.createTopic(topicName);
         } catch (JMSException e) {
             logger.error("Unable to create the topic with name " + topicName, e);
             throw new DispatcherException("Unable to create the topic with name " + topicName, e);
         }
     }

     protected abstract RemoteContext getRemoteContext(String alias) throws DispatcherException;

     protected JMSProvider getJMSProvider() {
         return jmsProvider;
     }

}
