/**
 * WildCAT: A Generic Framework for Context-Aware Applications.
 * Copyright (C) 2008 Bull S.A.S.
 * Copyright (C) 2008 EMN
 * Contact: wildcat@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * --------------------------------------------------------------------------
 * $Id: DispatcherFactory.java 355 2008-11-14 09:03:51Z loris $
 * --------------------------------------------------------------------------
 */
package org.ow2.wildcat.remote.dispatcher;

import java.net.URL;

import org.ow2.carol.util.configuration.CarolDefaultValues;
import org.ow2.carol.util.configuration.ConfigurationRepository;
import org.ow2.carol.util.configuration.ServerConfiguration;
import org.ow2.cmi.controller.common.ClusterViewManager;
import org.ow2.cmi.controller.factory.ClusterViewManagerFactory;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.wildcat.ContextFactory;
import org.ow2.wildcat.remote.dispatcher.jms.JMSDispatcher;
import org.ow2.wildcat.remote.dispatcher.rmi.RMIDispatcher;
import org.ow2.wildcat.remote.provider.JMSProvider;
import org.ow2.wildcat.remote.provider.JMSProviderException;

/**
 * Factory to create a single instance of {@link IDispatcher}.
 * Dispatchers allow contexts to access to some remote contexts and listen them.
 * @see RMIDispatcher an implementation (this used by default) using RMI to pull and JMS to push.
 * @see JMSDispatcher an implementation using JMS to pull and push
 * @author Loris Bouzonnet
 */
public final class DispatcherFactory {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(DispatcherFactory.class);

    /**
     * Singleton.
     */
    private static IDispatcher dispatcher;

    /**
     * No public constructor.
     */
    private DispatcherFactory() {
        // Nothing
    }

    /**
     * Return a dispatcher of accesses to remote contexts.
     * @param contextFactory a context factory to retrieve the configuration
     * @return a dispatcher of accesses to remote contexts
     * @throws DispatcherException if a dispatcher cannot be created
     */
    public static IDispatcher getDispatcher(final ContextFactory contextFactory) throws DispatcherException {
        if (dispatcher == null) {

            boolean useCmi = !Boolean.parseBoolean(contextFactory.getProperty(ContextFactory.KEY_PURE_JMS));
            boolean cmiStarted = false;
            ServerConfiguration serverConfig;

            System.setProperty(CarolDefaultValues.SERVER_MODE, "true");
            try {
                ConfigurationRepository.init();
                serverConfig = ConfigurationRepository.getServerConfiguration();
            } catch (Exception e) {
                logger.error("Unable to init Carol!", e);
                throw new DispatcherException("Unable to init Carol!", e);
            }

            JMSProvider delegate = contextFactory.getJMSProviderDelegate();
            if (delegate != null) {
                try {
                    delegate.init(contextFactory);
                } catch (JMSProviderException e) {
                    logger.error("Unable to init the JMS provider delegate", e);
                    throw new DispatcherException("Unable to init the JMS provider delegate", e);
                }
            }

            ClusterViewManager clusterViewManager = null;
            // Check if CMI is configured
            if (useCmi && !serverConfig.isStartCMI()) {
                URL url =
                    Thread.currentThread().getContextClassLoader().getResource(
                            contextFactory.getProperty(ContextFactory.KEY_CMI_CONFIG_FILENAME));
                try {
                    if (!serverConfig.isStartCMI()) {
                        serverConfig.enableCMI(url);
                    }
                    cmiStarted = true;
                    clusterViewManager = ClusterViewManagerFactory.getFactory().create();
                    clusterViewManager.start();
                } catch (Exception e) {
                    logger.info("Unable to configure CMI", e);
                    useCmi = false;
                }
            } else if (useCmi) {
                clusterViewManager = ClusterViewManagerFactory.getFactory().getClusterViewManager();
            }

            // Check if the CMI server is available
            if (useCmi && clusterViewManager != null && clusterViewManager.getState().equals(
                    ClusterViewManager.State.AVAILABLE)) {
                dispatcher = new RMIDispatcher(contextFactory);
            } else {
                if (cmiStarted) {
                    try {
                        if (clusterViewManager != null && serverConfig.isStartCMI()) {
                            serverConfig.disableCMI();
                            clusterViewManager.destroy();
                        }
                    } catch (Exception e) {
                        logger.debug("Error while disabling CMI", e);
                    }
                }
                // Else use pure JMS mode
                logger.info("Unable to use the RMI mode, using the pure JMS mode");
                dispatcher = new JMSDispatcher(contextFactory);
            }
        }
        return dispatcher;
    }

}
