/**
 * WildCAT: A Generic Framework for Context-Aware Applications.
 * Copyright (C) 2008 Bull S.A.S.
 * Copyright (C) 2008 EMN
 * Contact: wildcat@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors:
 *		- Pierre-Charles David (initial)
 *		- Nicolas Loriant
 * --------------------------------------------------------------------------
 * $Id: JMSDispatcher.java 350 2008-11-10 09:41:14Z loris $
 * --------------------------------------------------------------------------
 */
package org.ow2.wildcat.remote.dispatcher.jms;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import javax.jms.DeliveryMode;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueReceiver;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Session;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.wildcat.Context;
import org.ow2.wildcat.ContextFactory;
import org.ow2.wildcat.remote.dispatcher.AbstractDispatcher;
import org.ow2.wildcat.remote.dispatcher.DispatcherException;

/**
 * The Dispatcher class manage JMS communication between Context.
 * A Dispatcher instance manage multiple local context.
 * Notice that implementation doesn't yet work with JORAM.
 */
public class JMSDispatcher extends AbstractDispatcher {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(JMSDispatcher.class);

    private final String name;

    private final Map<Long, ContextOperationResult> pendingRemoteContextOperation;

    private long id = 0;

    private final QueueSession sessionToReceiveRequests;

    private final QueueSession sessionToSendRequests;

    private final QueueSession sessionToReceiveResponses;

    private final QueueSession sessionToSendResponses;

    /**
     * Receive response of pulls on remote context.
     */
    private final QueueReceiver responseQueueReceiver;

    private final Map<String, RemoteContextOperationListener> listeners;

    /**
     * Constructor
     *
     * @param factory
     * @param name
     * @throws Exception
     */
    public JMSDispatcher(final ContextFactory factory) throws DispatcherException {
        super(factory, false);
        name = factory.getProperty(ContextFactory.KEY_DISPATCHER_NAME);
        this.pendingRemoteContextOperation = new HashMap<Long, ContextOperationResult>();
        listeners = new HashMap<String, RemoteContextOperationListener>();

        try {
            QueueConnectionFactory qcf =
                getJMSProvider().getQueueConnectionFactory();

            QueueConnection qc = qcf.createQueueConnection();
            qc.start();
            sessionToReceiveRequests = qc.createQueueSession(false,
                    Session.AUTO_ACKNOWLEDGE);
            sessionToReceiveResponses = qc.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);
            sessionToSendRequests = qc.createQueueSession(false,
                    Session.AUTO_ACKNOWLEDGE);
            sessionToSendResponses = qc.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);
            Queue rq = getJMSProvider().getQueue(name);
            responseQueueReceiver = this.sessionToReceiveResponses.createReceiver(rq);
            responseQueueReceiver.setMessageListener(new ContextOperationListener());
        } catch (Exception e) {
            logger.error("Failed connecting JMS registry", e);
            throw new DispatcherException("Failed connecting JMS registry", e);
        }
    }

    /**
     * Makes a context callable remotely.
     * @param context a context
     * @throws DispatcherException
     */
    @Override
    public void doExport(final Context context) throws DispatcherException {
        try {
            Queue q = getJMSProvider().getQueue(context.getName() + "_requests");
            QueueReceiver qr = sessionToReceiveRequests.createReceiver(q);
            qr.setMessageListener(new ContextOperationListener());
            RemoteContextOperationListener listener = new RemoteContextOperationListener(q, qr);
            listeners.put(context.getName(), listener);
        } catch (Exception e) {
            logger.error("Unable to export the context with name " + context.getName(), e);
            throw new DispatcherException(
                    "Unable to export the context with name " + context.getName(), e);
        }
    }

    @Override
    protected JMSRemoteContext getRemoteContext(final String alias) {
        return new JMSRemoteContext(this, alias);
    }

    /**
     * Executes a remote call
     *
     * @param co the operation to perform
     * @return operation result
     * @throws DispatcherException
     */
    public ContextOperationResult doRemoteContextOperation(final ContextOperation co) throws DispatcherException {
        try {
            long coId = this.getNewId();
            co.setId(coId);
            co.setReturnQueue(name);
            Queue cq = sessionToSendRequests.createQueue(co.getTarget() + "_requests");
            QueueSender cqs = sessionToSendRequests.createSender(cq);
            cqs.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
            Message c = sessionToSendRequests.createObjectMessage(co);
            cqs.send(c);

            this.pendingRemoteContextOperation.put(coId, null);
            while (this.pendingRemoteContextOperation.get(coId) == null) {
                try {
                    Thread.sleep(10);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            }
            return this.pendingRemoteContextOperation.remove(coId);
        } catch (JMSException e) {
            logger.error("Unable to perform a remote invocation with the context operation: " + co, e);
            throw new DispatcherException(
                    "Unable to perform a remote invocation with the context operation: " + co, e);
        }
    }

    private long getNewId() {
        return this.id++ % Long.MAX_VALUE;
    }

    private class ContextOperationListener implements MessageListener {

        public void onMessage(final Message msg) {
            if (msg instanceof ObjectMessage) {
                Object obj = null;
                try {
                    obj = ((ObjectMessage) msg).getObject();
                    if (obj instanceof ContextOperation) {
                        ContextOperation co = (ContextOperation) obj;
                        Context callee = getExportedContext(co.getTarget());
                        Method method = null;
                        switch (co.getOp()) {
                        case GET_VALUE:
                            method = callee.getClass().getMethod("getValue",
                                    String.class);
                            break;
                        case SET_VALUE:
                            method = callee.getClass().getMethod("setValue",
                                    String.class, Object.class);
                            break;
                        case CREATE_ATTRIBUTE:
                            method = callee.getClass().getMethod("createAttribute",
                                    String.class, Object.class);
                            break;
                        case CREATE_RESOURCE:
                            method = callee.getClass().getMethod("createResource",
                                    String.class);
                            break;
                        case CREATE_SYMLINK:
                            method = callee.getClass().getMethod("createSymlink",
                                    String.class, String.class);
                            break;
                        case LIST:
                            method = callee.getClass().getMethod("list",
                                    String.class);
                            break;
                        }
                        ContextOperationResult cor = new ContextOperationResult();
                        cor.setId(co.getId());
                        cor.setResult(method.invoke(callee, co.getArgs()));

                        Queue rq = sessionToSendResponses.createQueue(co.getReturnQueue());
                        QueueSender rqs = sessionToSendResponses.createSender(rq);
                        rqs.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
                        Message r = sessionToSendResponses.createObjectMessage(cor);
                        rqs.send(r);
                        return;

                    } else if (obj instanceof ContextOperationResult) {
                        ContextOperationResult cor = (ContextOperationResult) obj;
                        pendingRemoteContextOperation.put(cor.getId(), cor);
                        return;
                    }

                } catch (Exception e) {
                    logger.warn("Error when processing a received message: " + msg, e);
                }
            }
            logger.warn("Unexpected message type");
        }
    }

    private class RemoteContextOperationListener {

        private final Queue queue;

        private final QueueReceiver receiver;

        RemoteContextOperationListener(final Queue queue, final QueueReceiver receiver) {
            this.queue = queue;
            this.receiver = receiver;
        }

    }

    /**
     * @param context
     * @throws DispatcherException
     */
    @Override
    protected void doUnexport(final Context context) throws DispatcherException {
        RemoteContextOperationListener listener =
            listeners.remove(context.getName());
        if(listener != null) {
            try {
                listener.receiver.close();
                getJMSProvider().deleteQueue(listener.queue.getQueueName());
            } catch (Exception e) {
                logger.error("Error while unexporting the context with name " + context.getName(), e);
                throw new DispatcherException(
                        "Error while unexporting the context with name" + context.getName(), e);
            }
        }
    }

}
